@@ -33,17 +33,20 @@ const Params = imports.params;
 const CategoryList = imports.categoryList;
 const Character = imports.character;
 const CharacterList = imports.characterList;
-const Pango = imports.gi.Pango;
-const Gc = imports.gi.Gc;
+const Menu = imports.menu;
 const Gettext = imports.gettext;
 
+const Main = imports.main;
 const Util = imports.util;
 
-const MAX_SEARCH_RESULTS = 100;
-
 const MainWindow = new Lang.Class({
     Name: 'MainWindow',
     Extends: Gtk.ApplicationWindow,
+    Template: 'resource:///org/gnome/Characters/mainwindow.ui',
+    InternalChildren: ['main-headerbar', 'search-active-button',
+                       'search-bar', 'search-entry',
+                       'menu-button',
+                       'main-grid', 'main-hbox', 'sidebar-grid'],
     Properties: {
         'search-active': GObject.ParamSpec.boolean(
             'search-active', '', '',
@@ -58,6 +61,7 @@ const MainWindow = new Lang.Class({
 
         this._searchActive = false;
         this._searchKeywords = [];
+        this._filterFontFamily = null;
 
         Util.initActions(this,
                          [{ name: 'about',
@@ -72,41 +76,44 @@ const MainWindow = new Lang.Class({
                             state: new GLib.Variant('s', 'punctuation') },
                           { name: 'character',
                             activate: this._character,
+                            parameter_type: new GLib.VariantType('s') },
+                          { name: 'filter-font',
+                            activate: this._filterFont,
                             parameter_type: new GLib.VariantType('s') }]);
 
-        let builder = new Gtk.Builder();
-        builder.add_from_resource('/org/gnome/Characters/mainwindow.ui');
-
-        this._headerBar = builder.get_object('main-headerbar');
-        this.set_titlebar(this._headerBar);
-
-        let searchBtn = builder.get_object('search-active-button');
-        this.bind_property('search-active', searchBtn, 'active',
+        this.bind_property('search-active', this._search_active_button, 'active',
                            GObject.BindingFlags.SYNC_CREATE |
                            GObject.BindingFlags.BIDIRECTIONAL);
-        this._searchBar = builder.get_object('search-bar');
         this.bind_property('search-active',
-                           this._searchBar,
+                           this._search_bar,
                            'search-mode-enabled',
                            GObject.BindingFlags.SYNC_CREATE |
                            GObject.BindingFlags.BIDIRECTIONAL);
-        let searchEntry = builder.get_object('search-entry');
-        this._searchBar.connect_entry(searchEntry);
-        searchEntry.connect('search-changed',
-                            Lang.bind(this, this._handleSearchChanged));
+        this._search_bar.connect_entry(this._search_entry);
+        this._search_entry.connect('search-changed',
+                                   Lang.bind(this, this._handleSearchChanged));
 
-        let grid = builder.get_object('main-grid');
+        this._menu_popover = new Menu.MenuPopover({});
+        this._menu_button.set_popover(this._menu_popover);
 
-        let sidebarGrid = builder.get_object('sidebar-grid');
-        this._categoryList = new CategoryList.CategoryListWidget();
-        sidebarGrid.add(this._categoryList);
+        this._categoryList =
+            new CategoryList.CategoryListWidget({ vexpand: true });
+        let scroll = new Gtk.ScrolledWindow({
+            hscrollbar_policy: Gtk.PolicyType.NEVER,
+            hexpand: false,
+        });
+        scroll.add(this._categoryList);
+        this._sidebar_grid.add(scroll);
 
-        let hbox = builder.get_object('main-hbox');
         this._mainView = new MainView();
-        hbox.pack_start(this._mainView, true, true, 0);
 
-        this.add(grid);
-        grid.show_all();
+        if (this._mainView.recentCharacters.length == 0) {
+            let row = this._categoryList.get_row_at_index(1);
+            this._categoryList.select_row(row);
+        }
+
+        this._main_hbox.pack_start(this._mainView, true, true, 0);
+        this._main_grid.show_all();
 
         // Due to limitations of gobject-introspection wrt GdkEvent
         // and GdkEventKey, this needs to be a signal handler
@@ -121,6 +128,7 @@ const MainWindow = new Lang.Class({
         if (this._searchActive == v)
             return;
 
+        this._categoryList.unselect_all();
         this._searchActive = v;
         this.notify('search-active');
     },
@@ -132,7 +140,7 @@ const MainWindow = new Lang.Class({
         if (keywords != this._searchKeywords) {
             this._searchKeywords = keywords;
             if (this._searchKeywords.length > 0)
-                this._mainView.startSearch(this._searchKeywords);
+                this._mainView.searchByKeywords(this._searchKeywords);
             else
                 this._mainView.cancelSearch();
         }
@@ -140,7 +148,9 @@ const MainWindow = new Lang.Class({
     },
 
     _handleKeyPress: function(self, event) {
-        return this._searchBar.handle_event(event);
+        if (this._menu_popover.visible)
+            return false;
+        return this._search_bar.handle_event(event);
     },
 
     _about: function() {
@@ -149,11 +159,11 @@ const MainWindow = new Lang.Class({
               translator_credits: _("translator-credits"),
               program_name: _("GNOME Characters"),
               comments: _("Character Map"),
-              copyright: 'Copyright 2014 Daiki Ueno',
+              copyright: 'Copyright 2014-2015 Daiki Ueno',
               license_type: Gtk.License.GPL_2_0,
               logo_icon_name: 'org.gnome.Characters',
               version: pkg.version,
-              website: 'http://www.example.com/gnome-characters/',
+              // website: 'https://wiki.gnome.org/Design/Apps/CharacterMap',
               wrap_license: true,
               modal: true,
               transient_for: this
@@ -165,8 +175,21 @@ const MainWindow = new Lang.Class({
         });
     },
 
+    _updateTitle: function(title) {
+        if (this._filterFontFamily) {
+            this._main_headerbar.title =
+                _("%s (%s only)").format(Gettext.gettext(title),
+                                         this._filterFontFamily);
+        } else {
+            this._main_headerbar.title = Gettext.gettext(title);
+        }
+    },
+
     _category: function(action, v) {
         let [name, length] = v.get_string()
+
+        // FIXME: we could use Gtk.Container.get_child to obtain the
+        // title, but it is not introspectable.
         let category = null;
         for (let index in CategoryList.Category) {
             category = CategoryList.Category[index];
@@ -175,153 +198,113 @@ const MainWindow = new Lang.Class({
         }
 
         Util.assertNotEqual(category, null);
-        this._mainView.setPage(category.name);
-        this._headerBar.title = Gettext.gettext(category.label);
+        this._mainView.setPage(category.name, this._filterFontFamily);
+        this._updateTitle(category.title);
     },
 
     _character: function(action, v) {
         let [uc, length] = v.get_string()
-        this._mainView.selectCharacter(uc);
+        this._mainView.addToRecent(uc);
+    },
+
+    _filterFont: function(action, v) {
+        let [family, length] = v.get_string()
+        if (family == 'None')
+            family = null;
+        this._mainView.setFilterFont(family);
+        this._filterFontFamily = family;
+        this._updateTitle(this._mainView.visible_child.title);
     },
 });
 
 const MainView = new Lang.Class({
     Name: 'MainView',
     Extends: Gtk.Stack,
+    Properties: {
+        'max-recent-characters': GObject.ParamSpec.uint(
+            'max-recent-characters', '', '',
+            GObject.ParamFlags.READABLE | GObject.ParamFlags.WRITABLE,
+            0, GLib.MAXUINT32, 100)
+    },
+
+    get max_recent_characters() {
+        return this._maxRecentCharacters;
+    },
+
+    set max_recent_characters(v) {
+        this._maxRecentCharacters = v;
+        if (this.recentCharacters.length > this._maxRecentCharacters)
+            this.recentCharacters = this.recentCharacters.slice(
+                0, this._maxRecentCharacters);
+    },
 
     _init: function(params) {
         params = Params.fill(params, { hexpand: true, vexpand: true });
         this.parent(params);
 
-        this._characterListWidgets = {};
+        this._characterLists = {};
 
         let characterList;
         for (let index in CategoryList.Category) {
             let category = CategoryList.Category[index];
-            characterList = this._createCharacterList();
-            characterList.get_accessible().accessible_name =
-                _('%s Character List').format(category.label);
-            this._characterListWidgets[category.name] = characterList;
-            this.add_named(this._createScrolledWindow(characterList),
-                           category.name);
+            characterList = this._createCharacterList(
+                category.name, _('%s Character List').format(category.title));
+            // FIXME: Can't use GtkContainer.child_get_property.
+            characterList.title = category.title;
+            this.add_titled(characterList, category.name, category.title);
         }
 
-        characterList = this._createCharacterList();
-        characterList.get_accessible().accessible_name =
-            _('Search Result Character List');
-        this.add_named(this._createScrolledWindow(characterList),
-                       'search-result');
-        this._characterListWidgets['search-result'] = characterList;
+        characterList = this._createCharacterList(
+            'search-result', _('Search Result Character List'));
+        // FIXME: Can't use GtkContainer.child_get_property.
+        characterList.title = "Search Result";
+        this.add_named(characterList, 'search-result');
 
-        let builder = new Gtk.Builder();
-        builder.add_from_resource('/org/gnome/Characters/mainview.ui');
-        let searchBannerGrid = builder.get_object('search-banner-grid');
-        this.add_named(searchBannerGrid, 'search-banner');
-        let loadingBannerGrid = builder.get_object('loading-banner-grid');
-        this._spinner = builder.get_object('loading-banner-spinner');
-        this.add_named(loadingBannerGrid, 'loading-banner');
-        this._spinnerTimeoutId = 0;
-
-        this._recentCharacters = [];
-        this._cancellable = new Gio.Cancellable();
+        // FIXME: Can't use GSettings.bind with 'as' from Gjs
+        let recentCharacters = Main.settings.get_value('recent-characters');
+        this.recentCharacters = recentCharacters.get_strv();
+        this._maxRecentCharacters = 100;
+        Main.settings.bind('max-recent-characters', this,
+                           'max-recent-characters',
+                           Gio.SettingsBindFlags.DEFAULT);
     },
 
-    _createCharacterList: function() {
-        let widget = new CharacterList.CharacterListWidget({ hexpand: true,
-                                                             vexpand: true });
+    _createCharacterList: function(name, accessible_name) {
+        let characterList = new CharacterList.CharacterListView({});
+        characterList.get_accessible().accessible_name = accessible_name;
+
+        let scroll = characterList.get_child_by_name('character-list');
+        let widget = scroll.get_child().get_child();
         widget.connect('character-selected',
                        Lang.bind(this, this._handleCharacterSelected));
-        return widget;
+
+        this._characterLists[name] = characterList;
+        return characterList;
     },
 
-    _createScrolledWindow: function(widget) {
-        let viewport = new Gtk.Viewport({});
-        viewport.add(widget);
-
-        let scroll = new Gtk.ScrolledWindow({
-            hscrollbar_policy: Gtk.PolicyType.NEVER
-        });
-        scroll.add(viewport);
-        return scroll;
-    },
-
-    _startSearch: function() {
-        this._cancellable.cancel();
-        this._cancellable.reset();
-
-        if (this.visible_child_name != 'search-banner' &&
-            this.visible_child_name != 'search-result')
-            this._lastPage = this.visible_child_name;
-
-        this._spinnerTimeoutId =
-            GLib.timeout_add(GLib.PRIORITY_DEFAULT, 1000,
-                             Lang.bind(this, function () {
-                                 this._spinner.start();
-                                 this.visible_child_name = 'loading-banner';
-                                 this.show_all();
-                             }));
-    },
-
-    _finishSearch: function(name, result) {
-        if (this._spinnerTimeoutId > 0) {
-            GLib.source_remove(this._spinnerTimeoutId);
-            this._spinnerTimeoutId = 0;
-            this._spinner.stop();
-        }
-
-        let characters = [];
-        for (let index = 0; index < result.len; index++) {
-            characters.push(Gc.search_result_get(result, index));
-        }
-
-        let characterList = this._characterListWidgets[name];
-        characterList.setCharacters(characters);
-        if (result.length == 0) {
-            if (this._lastPage)
-                this.visible_child_name = this._lastPage;
-            else
-                this.visible_child_name = 'search-banner';
-        } else {
-            this.visible_child_name = name;
-        }
-        this.show_all();
-    },
-
-    startSearch: function(keywords) {
-        this._startSearch();
-        Gc.search_by_keywords(
-            keywords,
-            MAX_SEARCH_RESULTS,
-            this._cancellable,
-            Lang.bind(this, function(source_object, res, user_data) {
-                try {
-                    let result = Gc.search_finish(res);
-                    this._finishSearch('search-result', result);
-                } catch (e) {
-                    log("Failed to search by keywords: " + e);
-                }
-            }));
+    searchByKeywords: function(keywords) {
+        this.visible_child_name = 'search-result';
+        this.visible_child.searchByKeywords(keywords);
     },
 
     cancelSearch: function() {
-        this._cancellable.cancel();
-        this._finishSearch('search-result', []);
+        this.visible_child.cancelSearch();
     },
 
-    setPage: function(name) {
-        if (!(name in this._characterListWidgets))
+    setPage: function(name, filterFontFamily) {
+        if (!(name in this._characterLists))
             return;
 
-        let characterList = this._characterListWidgets[name];
+        this.visible_child_name = name;
+        this.visible_child.setFilterFont(filterFontFamily);
+
         if (name == 'recent') {
-            if (this._recentCharacters.length == 0)
-                this.visible_child_name = 'search-banner';
+            if (this.recentCharacters.length == 0)
+                this.visible_child.visible_child_name = 'empty-recent';
             else {
-                characterList.setCharacters(this._recentCharacters);
-                characterList.show_all();
-                this.visible_child_name = name;
+                this.visible_child.setCharacters(this.recentCharacters);
+                this.visible_child.updateCharacterList();
             }
-            this.show_all();
         } else {
             let category = null;
             for (let index in CategoryList.Category) {
@@ -331,39 +314,30 @@ const MainView = new Lang.Class({
             }
 
             Util.assertNotEqual(category, null);
-            this._startSearch();
-            Gc.search_by_category(
-                category.category,
-                -1,
-                this._cancellable,
-                Lang.bind(this,
-                          function(source_object, res, user_data) {
-                              try {
-                                  let result = Gc.search_finish(res);
-                                  this._finishSearch(name, result);
-                              } catch (e) {
-                                  log("Failed to search by category: " + e);
-                              }
-                          }));
+            this.visible_child.searchByCategory(category);
         }
     },
 
-    selectCharacter: function(uc) {
-        if (this._recentCharacters.indexOf(uc) < 0) {
-            this._recentCharacters.push(uc);
-
-            if (this.visible_child_name == 'recent')
-                this.setPage('recent');
+    addToRecent: function(uc) {
+        if (this.recentCharacters.indexOf(uc) < 0) {
+            this.recentCharacters.push(uc);
+            if (this.recentCharacters.length > this._maxRecentCharacters)
+                this.recentCharacters = this.recentCharacters.slice(
+                    0, this._maxRecentCharacters);
+            Main.settings.set_value(
+                'recent-characters',
+                GLib.Variant.new_strv(this.recentCharacters));
         }
     },
 
     _handleCharacterSelected: function(widget, uc) {
-        this.selectCharacter(uc);
+        this.addToRecent(uc);
 
         let dialog = new Character.CharacterDialog({
             character: uc,
             modal: true,
-            transient_for: this.get_toplevel()
+            transient_for: this.get_toplevel(),
+            fontDescription: this.visible_child.getFontDescription()
         });
 
         dialog.show();
@@ -371,5 +345,9 @@ const MainView = new Lang.Class({
             if (response_id == Gtk.ResponseType.CLOSE)
                 dialog.destroy();
         });
+    },
+
+    setFilterFont: function(family) {
+        this.visible_child.setFilterFont(family);
     }
 });
